import pygame
from showml.simulations.conways_game_of_life.game_window import GameWindow
from showml.simulations.conways_game_of_life.event import Action, Event


class EventHandler:
    """The EventHandler class responsible for observing events taking place in the window as well as initializing and managing the Grid.
    """

    def __init__(self, window: GameWindow) -> None:
        """Constructor for the EventHandler class.

        Args:
            window (GameWindow): A game window object where the grid and buttons/text live.
        """
        self.window = window

    def get_event(self) -> Event:
        """This method returns an Event object based on the Action taken by the user.
        It checks the collide point of the user's mouse click with the different entities in the window.

        Returns:
            Event: An Event object containing the Action performed by the user (and also the row, column if a cell is toggled).
        """
        for event in pygame.event.get():
            x, y = pygame.mouse.get_pos()

            if event.type == pygame.QUIT:
                pygame.quit()

            elif event.type == pygame.MOUSEBUTTONDOWN:
                if x < self.window.SCREEN_WIDTH - 100 and y:
                    row = y // (self.window.CELL_HEIGHT + self.window.CELL_MARGIN)
                    column = x // (self.window.CELL_WIDTH + self.window.CELL_MARGIN)
                    return Event(action=Action.CELL_TOGGLE, row=row, column=column)
                else:
                    # Check if a button is pressed
                    for button in self.window.buttons:
                        if button.button.collidepoint(x, y):
                            return button.on_click()

        return Event(action=Action.NO_EVENT)



from showml.losses import MeanSquaredError
from showml.optimizers import SGD
import numpy as np
from numpy.testing import assert_almost_equal


def test_sgd_mse_update_weights() -> None:
    X = np.array([[0.5, 1], [-1, 1], [7, -6], [6, 8], [1, 1]])
    y = np.array([0.5, 1, -1, 1, -6])
    z = np.array([0, 2, -1, 2, -5])
    weights = np.array([1.4, 7.8])
    bias = [2.6]
    optimizer = SGD(learning_rate=0.01)
    loss = MeanSquaredError()
    dw, db = loss.parameter_gradient(X, y, z)
    assert_almost_equal(dw, [1.15, 1.9])
    assert db == 0.5
    updated_weights, updated_bias = optimizer.update_weights(weights, bias, dw, db)
    assert_almost_equal(updated_weights, [1.3885, 7.781])
    assert updated_bias == 2.595


def test_sgd_mse_update_weights_simple() -> None:
    X = np.array([[1.0, 2.0, 3.0, 4.0], [2.0, 1.0, 7.0, 8.0]])
    y = np.array([10.0, 18.0])
    z = np.array([5.0, 7.0])
    weights = np.array([1.0, 1.0, 1.0, 1.0])
    bias = [0.0]
    optimizer = SGD(learning_rate=0.1)
    loss = MeanSquaredError()
    dw, db = loss.parameter_gradient(X, y, z)
    assert_almost_equal(dw, [-13.5, -10.5, -46.0, -54.0])
    assert db == -8.0
    updated_weights, updated_bias = optimizer.update_weights(weights, bias, dw, db)
    assert_almost_equal(updated_weights, [2.35, 2.05, 5.6, 6.4])
    assert updated_bias == 0.8


def test_sgd_mse_update_weights_int() -> None:
    X = np.array([[1, 2, 3, 4], [2, 1, 7, 8]]).astype("float64")
    y = np.array([10, 18]).astype("float64")
    z = np.array([5, 7]).astype("float64")
    weights = np.array([1, 1, 1, 1]).astype("float64")
    bias = [0]
    optimizer = SGD(learning_rate=0.1)
    loss = MeanSquaredError()
    dw, db = loss.parameter_gradient(X, y, z)
    assert_almost_equal(dw, [-13.5, -10.5, -46.0, -54.0])
    assert db == -8.0
    updated_weights, updated_bias = optimizer.update_weights(weights, bias, dw, db)
    assert_almost_equal(updated_weights, [2.35, 2.05, 5.6, 6.4])
    assert updated_bias == 0.8



from abc import ABC, abstractmethod
from typing import Tuple
from showml.optimizers.base_optimizer import Optimizer
import numpy as np


class Layer(ABC):
    """A layer class.
    """

    def __init__(
        self, input_shape: Tuple[int] = (0,), has_weights: bool = True
    ) -> None:
        """Constructor for the Layer class.

        Args:
            input_shape (Tuple[int], optional): The shape of the input (This is set to the previous layer's output shape if it is NOT the first layer of the network). Defaults to (0,).
            has_weights (bool, optional): A boolean to indicate if this layer has trainable params. Activation layers have this flag set to False. Defaults to True.
        """
        self.input_shape = input_shape
        self.has_weights = has_weights

    @abstractmethod
    def initialize_params(self, optimizer: Optimizer) -> None:
        """Initializes the weights and bias of the layer.

        Args:
            optimizer (Optimizer): The optimizer to be used for training (showml.optimizers).
        """
        pass

    @abstractmethod
    def forward(self, X: np.ndarray) -> np.ndarray:
        """A method which computes a forward pass of a layer.

        Args:
            X (np.ndarray): The input to the layer.

        Returns:
            np.ndarray: Output of the layer (multiplying the weights and adding the bias term).
        """
        pass

    @abstractmethod
    def backward(self, X: np.ndarray) -> np.ndarray:
        """A method which computes a backward pass of a layer.

        Args:
            X (np.ndarray): Corresponds to the input to the layer in the backward propagation.

        Returns:
            np.ndarray: Output of the layer after backpropagation.
        """
        pass

    @abstractmethod
    def get_output_shape(self) -> Tuple[int]:
        """Returns the output shape of a layer.

        Returns:
            Tuple[int]: The output shape of the layer.
        """
        pass

    @abstractmethod
    def get_params_count(self) -> int:
        """Computes the number of trainable parameters of a layer.

        Returns:
            int: The number of trainable parameters of a layer.
        """
        pass



import numpy as np


class Grid:
    """A 2D Grid class to display the cells.
    """

    def __init__(self, num_rows: int, num_columns: int) -> None:
        """[summary]

        Args:
            num_rows (int): The number of rows in the grid.
            num_columns (int): The number of columns in the grid.
        """
        self.num_rows = num_rows
        self.num_columns = num_columns
        self.grid: np.ndarray = np.zeros((num_rows, num_columns))

        # Set initial live cells
        self.grid[9][11] = 1
        self.grid[10][11] = 1
        self.grid[11][11] = 1
        self.grid[10][12] = 1
        self.grid[10][10] = 1

    def calculate_neighbor_sum(self, row: int, column: int) -> int:
        """This method computes the sum of the 8-neighbors of the cell at [row, column].
        Reference: https://stackoverflow.com/a/37026344

        Args:
            row (int): The row index.
            column (int): The column index.

        Returns:
            int: Sum of the 8 neighbors of the cell.
        """
        # Extract the region (8-neighbors and self) of element at [row, column]
        region = self.grid[max(0, row - 1) : row + 2, max(0, column - 1) : column + 2]
        # Sum the region and subtract center
        return np.sum(region) - self.grid[row, column]

    def update_grid(self) -> None:
        """This method updates the grid using the rules of Conway's Game of Life.
        Reference: https://en.wikipedia.org/wiki/Conway%27s_Game_of_Life
        """
        new_grid = np.zeros((self.num_rows, self.num_columns))
        for row in range(self.num_rows):
            for column in range(self.num_columns):
                neighbor_sum = self.calculate_neighbor_sum(row, column)
                if self.grid[row][column] == 1:
                    if neighbor_sum == 2 or neighbor_sum == 3:
                        new_grid[row][column] = 1
                    else:
                        new_grid[row][column] = 0
                elif self.grid[row][column] == 0:
                    if neighbor_sum == 3:
                        new_grid[row][column] = 1
                    else:
                        new_grid[row][column] = 0
        self.grid = new_grid

    def reset_grid(self) -> None:
        """This method resets the grid i.e., kills all the cells.
        """
        self.grid = np.zeros((self.num_rows, self.num_columns))

    def toggle_cell_value(self, row: int, column: int) -> None:
        """This method toggles the value of a cell (sets to 1 if 0 else 0)

        Args:
            row (int): The row index.
            column (int): The column index.
        """
        self.grid[row][column] = 1 if self.grid[row][column] == 0 else 0



from typing import List
import matplotlib.pyplot as plt
import numpy as np

from showml.utils.exceptions import InvalidShapeError


def generic_metric_plot(metric_name: str, metric_values: List[float]) -> None:
    """Plot the metric values after training (epoch vs metric).

    Args:
        metric_name (str): Name of the metric (accuracy, r^2 score, etc.).
        metric_values (List[float]): A list of metric values collected during training for all epochs.
    """
    plt.plot(metric_values)
    plt.xlabel("Epoch")
    plt.ylabel(metric_name)
    plt.title(metric_name)
    plt.show()


def plot_regression_line(
    X: np.ndarray, y: np.ndarray, z: np.ndarray, xlabel: str = "", ylabel: str = ""
) -> None:
    """Plot the regression line to visualize how well the model fits to the data.
    Only works when the entire dataset is 2-dimensional i.e., input data (X) is 1-dimensional.

    Args:
        X (np.ndarray): The input data.
        y (np.ndarray): The true labels of the input data.
        z (np.ndarray): The predicted values for the input data.
        xlabel (str, optional): The label corresponding to the input feature name. Defaults to "".
        ylabel (str, optional): The label corresponding to the output feature name. Defaults to "".
    """
    if X.shape[1] != 1:
        raise InvalidShapeError("X must have exactly 1 dimension.")

    plt.scatter(X, y, color="red")
    plt.plot(X, z, color="blue")
    plt.title("Regression Line")
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.show()



from typing import Tuple
from abc import ABC, abstractmethod
import numpy as np


class Optimizer(ABC):
    """The Base Optimizer Class.
    """

    def __init__(self, learning_rate: float):
        """Constructor for the Base Optimzer class.

        Args:
            learning_rate (float): The learning rate (how much to update the weights at each iteration).
        """
        self.learning_rate = learning_rate

    @abstractmethod
    def update_weights(
        self, weights: np.ndarray, bias: np.ndarray, dw: np.ndarray, db: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Update the weights of the model using the specified loss function and optimizer.

        Args:
            weights (np.ndarray): The set of training weights of the model.
            bias (np.ndarray): The bias value(s) of the model.
            dw (np.ndarray): Gradient of the weights w.r.t. loss function.
            db (np.ndarray): Gradient of the bias w.r.t. loss function.

        Returns:
            np.ndarray: The set of updated weights after optimization for an epoch.
            np.ndarray: The set of updated bias value(s) after optimization for an epoch.
        """
        pass



from abc import ABC, abstractmethod
from typing import Tuple
import numpy as np


class Loss(ABC):
    @abstractmethod
    def objective(self, y: np.ndarray, z: np.ndarray) -> float:
        """The objective loss function.

        Args:
            y (np.ndarray): The true labels of the data.
            z (np.ndarray): The predicted labels.

        Returns:
            float: The loss value.
        """
        pass

    @abstractmethod
    def parameter_gradient(
        self, X: np.ndarray, y: np.ndarray, z: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Computes the derivative of the weights and bias w.r.t the objective function.

        Args:
            X (np.ndarray): The training data.
            y (np.ndarray): The true labels of the data.
            z (np.ndarray): The predicted labels.

        Returns:
            np.ndarray: Gradient of the loss function with respect to the weights.
            np.ndarray: Gradient of the loss function with respect to the bias.
        """
        pass

    @abstractmethod
    def objective_gradient(self, y: np.ndarray, z: np.ndarray) -> np.ndarray:
        """Computes the gradient of the objective function.

        Args:
            y (np.ndarray): The true values.
            z (np.ndarray): The predicted values.

        Returns:
            np.ndarray: Gradient of the objective function.
        """
        pass



from typing import Tuple
from showml.losses.base_loss import Loss
import numpy as np


class MeanSquaredError(Loss):
    def objective(self, y: np.ndarray, z: np.ndarray) -> float:
        return np.average(np.square(z - y), axis=0)

    def objective_gradient(self, y: np.ndarray, z: np.ndarray) -> np.ndarray:
        # TODO
        pass

    def parameter_gradient(
        self, X: np.ndarray, y: np.ndarray, z: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        error = z - y
        num_samples = len(error)
        dw = (1 / num_samples) * X.T.dot(error)
        db = (1 / num_samples) * np.sum(error)
        return dw, db


class BinaryCrossEntropy(Loss):
    def objective(self, y: np.ndarray, z: np.ndarray) -> float:
        """Also known as Log Loss
        """
        num_samples = len(y)
        # Avoid division by zero
        z = np.clip(z, 1e-15, 1 - 1e-15)
        return -(1 / num_samples) * (np.sum(y * np.log(z) + (1 - y) * np.log(1 - z)))

    def objective_gradient(self, y: np.ndarray, z: np.ndarray) -> np.ndarray:
        z = np.clip(z, 1e-15, 1 - 1e-15)
        return (-y / z) - (1 - y) / (z - 1)

    def parameter_gradient(
        self, X: np.ndarray, y: np.ndarray, z: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        error = z - y
        num_samples = len(error)
        dw = (1 / num_samples) * np.dot(X.T, (error))
        db = (1 / num_samples) * np.sum((error))
        return dw, db


class CrossEntropy(Loss):
    def objective(self, y: np.ndarray, z: np.ndarray) -> float:
        """param y: one hot encoded values
        """
        num_samples = len(z)
        z = np.clip(z, 1e-15, 1.0 - 1e-15)
        return -np.sum(y * np.log(z)) / num_samples

    def objective_gradient(self, y: np.ndarray, z: np.ndarray) -> np.ndarray:
        z = np.clip(z, 1e-15, 1 - 1e-15)
        return -(y / z) + (1 - y) / (1 - z)

    def parameter_gradient(
        self, X: np.ndarray, y: np.ndarray, z: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        # TODO
        pass



from abc import ABC, abstractmethod
from typing import Tuple
from showml.simulations.conways_game_of_life.config import Color
from showml.simulations.conways_game_of_life.event import Action, Event

import pygame


class Button(ABC):
    def __init__(
        self,
        text: str,
        text_x: int,
        text_y: int,
        width: int,
        height: int,
        x: int,
        y: int,
        color: Tuple[int, int, int],
        screen: pygame.surface.Surface,
    ):
        self.text = text
        self.text_x = text_x
        self.text_y = text_y
        self.width = width
        self.height = height
        self.x = x
        self.y = y
        self.color = color
        self.button = pygame.draw.rect(screen, color, (x, y, width, height))

    @abstractmethod
    def on_click(self) -> Event:
        pass


class StartButton(Button):
    def __init__(self, screen: pygame.surface.Surface, color: Tuple[int, int, int]):
        screen_info = screen.get_rect()
        w: int = screen_info[2]
        super().__init__(
            text="Start",
            text_x=w - 65,
            text_y=13,
            width=60,
            height=20,
            x=w - 80,
            y=10,
            color=color,
            screen=screen,
        )

    def on_click(self) -> Event:
        return Event(action=Action.START)


class StopButton(Button):
    def __init__(self, screen: pygame.surface.Surface, color: Tuple[int, int, int]):
        screen_info = screen.get_rect()
        w: int = screen_info[2]
        super().__init__(
            text="Stop",
            text_x=w - 65,
            text_y=42,
            width=60,
            height=20,
            x=w - 80,
            y=40,
            color=color,
            screen=screen,
        )

    def on_click(self) -> Event:
        return Event(action=Action.STOP)


class ResetButton(Button):
    def __init__(self, screen: pygame.surface.Surface, color: Tuple[int, int, int]):
        screen_info = screen.get_rect()
        w: int = screen_info[2]
        super().__init__(
            text="Reset",
            text_x=w - 68,
            text_y=72,
            width=60,
            height=20,
            x=w - 80,
            y=70,
            color=color,
            screen=screen,
        )

    def on_click(self) -> Event:
        return Event(action=Action.RESET)



from typing import Tuple
import numpy as np
from showml.deep_learning.base_layer import Layer
from showml.optimizers.base_optimizer import Optimizer


class Dense(Layer):
    """A Dense Layer.
    """

    def __init__(self, num_nodes: int, input_shape: Tuple[int] = (0,)):
        """Constructor for the Dense Layer.

        Args:
            num_nodes (int): Initializes a Dense layer with the specified number of neurons.
            input_shape (Tuple[int], optional): The number of neurons in the layer. Defaults to (0,).
        """
        self.num_nodes = num_nodes
        super().__init__(input_shape=input_shape)

    def initialize_params(self, optimizer: Optimizer) -> None:
        self.optimizer = optimizer
        limit = 1 / np.sqrt(self.num_nodes)
        self.weights = np.random.uniform(
            -limit, limit, (self.input_shape[0], self.num_nodes)
        )
        self.bias = np.zeros((1, self.num_nodes))

    def get_params_count(self) -> int:
        return np.prod(self.weights.shape) + np.prod(self.bias.shape)

    def get_output_shape(self) -> Tuple[int]:
        return (self.num_nodes,)

    def forward(self, X: np.ndarray) -> np.ndarray:
        self.layer_input = X
        return X.dot(self.weights) + self.bias

    def backward(self, X: np.ndarray) -> np.ndarray:
        old_weights = self.weights

        dw = self.layer_input.T.dot(X)
        db = np.sum(X, axis=0, keepdims=True)

        self.weights, self.bias = self.optimizer.update_weights(
            self.weights, self.bias, dw, db
        )

        return X.dot(old_weights.T)


class Activation(Layer):
    def __init__(self, input_shape: Tuple[int] = (0,)):
        super().__init__(input_shape=input_shape, has_weights=False)

    def get_output_shape(self) -> Tuple[int]:
        return self.input_shape

    def get_params_count(self) -> int:
        return 0

    def initialize_params(self, optimizer: Optimizer) -> None:
        pass



from typing import List, Tuple
import pygame
from showml.simulations.conways_game_of_life.button import (
    Button,
    StartButton,
    StopButton,
    ResetButton,
)
from showml.simulations.conways_game_of_life.config import GameWindowSettings, Color
from showml.simulations.conways_game_of_life.grid import Grid


class GameWindow:
    """The GameWindow class responsible for observing events taking place in the window as well as initializing and managing the Grid.
    """

    def __init__(self, grid: Grid) -> None:
        """Constructor for the GameWindow class.

        Args:
            grid (Grid): A 2D grid containing cells where the simulation will take place.
        """
        self.window_settings = GameWindowSettings()

        pygame.init()
        pygame.display.set_caption(self.window_settings.CAPTION)

        self.grid = grid

        self.CELL_HEIGHT = self.window_settings.CELL_HEIGHT
        self.CELL_WIDTH = self.window_settings.CELL_WIDTH
        self.CELL_MARGIN = self.window_settings.CELL_MARGIN

        self.SCREEN_WIDTH, self.SCREEN_HEIGHT = self._compute_screen_width_and_height()
        self.SCREEN = pygame.display.set_mode((self.SCREEN_WIDTH, self.SCREEN_HEIGHT))

        self.buttons: List[Button] = [
            StartButton(self.SCREEN, Color.WHITE),
            StopButton(self.SCREEN, Color.WHITE),
            ResetButton(self.SCREEN, Color.WHITE),
        ]

        self.clock = pygame.time.Clock()

    def display_window_and_grid(self, delay: int) -> None:
        """This method is repsonsible for displaying the entire Game window with the grid, buttons and textual entities.

        Args:
            delay (int): The delay in milliseconds between each iteration.
        """
        pygame.time.wait(delay)
        self._display_buttons_and_text()
        self._display_grid()
        pygame.display.flip()
        self.clock.tick(60)

    def _display_buttons_and_text(self) -> None:
        """This private method displays the buttons and the text objects in the window.
        """
        self.SCREEN.fill(Color.BLACK)
        for button in self.buttons:
            # Button
            pygame.draw.rect(
                self.SCREEN,
                button.color,
                (button.x, button.y, button.width, button.height),
            )

            # Button Text
            self.SCREEN.blit(
                pygame.font.SysFont(
                    self.window_settings.FONT, self.window_settings.FONT_SIZE
                ).render(button.text, True, Color.BLACK),
                (button.text_x, button.text_y),
            )

    def _display_grid(self) -> None:
        """This private method displays the entire grid in the window.
        """
        for row in range(self.grid.num_rows):
            for column in range(self.grid.num_columns):
                if self.grid.grid[row][column] == 1:
                    # Alive cell
                    color = Color.WHITE
                else:
                    # Dead cell
                    color = Color.GRAY
                pygame.draw.rect(
                    self.SCREEN,
                    color,
                    [
                        self.window_settings.CELL_MARGIN
                        + (
                            self.window_settings.CELL_MARGIN
                            + self.window_settings.CELL_WIDTH
                        )
                        * column,
                        self.window_settings.CELL_MARGIN
                        + (
                            self.window_settings.CELL_MARGIN
                            + self.window_settings.CELL_HEIGHT
                        )
                        * row,
                        self.window_settings.CELL_WIDTH,
                        self.window_settings.CELL_HEIGHT,
                    ],
                )

    def _compute_screen_width_and_height(self) -> Tuple[int, int]:
        """This method computes the width and height of the screen based on the defined cell size and number of rows and columns.

        Returns:
            int: The screen width
            int: The screen height
        """
        screen_width = (
            self.grid.num_columns * self.CELL_WIDTH + self.grid.num_columns + 100
        )
        screen_height = self.grid.num_rows * self.CELL_HEIGHT + self.grid.num_rows
        return screen_width, screen_height



from typing import Callable, Dict, List
import numpy as np
from terminaltables import AsciiTable
from showml.losses.base_loss import Loss
from showml.optimizers.base_optimizer import Optimizer
from showml.deep_learning.layers import Layer
from showml.utils.dataset import Dataset
from showml.utils.model_utilities import generate_minibatches
import copy


class Sequential:
    """A Sequential model (deep neural network) with various types of layers and activation functions.
    """

    def __init__(self) -> None:
        """Constructor for the Sequential model class.
        """
        self.layers: List[Layer] = []

    def compile(
        self, optimizer: Optimizer, loss: Loss, metrics: List[Callable] = []
    ) -> None:
        """Compiles the model with the specified optimizer and evaluation metrics.
        This method also initializes the model.history object to store metric values during training.
        This method also initializes all the layers.

        Args:
            optimizer (Optimizer): The optimizer to be used for training (showml.optimizers).
            loss (Loss): The loss function used by the model to evaluate the solution.
            metrics (List[Callable], optional): A list of metrics which have to be calculated and displayed for model evaluation. Defaults to [].
        """
        self.optimizer = optimizer
        self.loss = loss
        self.metrics = metrics
        self.history: Dict[str, List[float]] = {
            metric.__name__: [] for metric in self.metrics
        }
        self.initialize_layers()

    def initialize_layers(self) -> None:
        """Initializes all the layers with the specified optimizer and parameters.
        """
        for layer_idx, layer in enumerate(self.layers):
            if layer_idx > 0:
                # If this layer is NOT the first layer of the network, then input shape = output shape of previous layer
                layer.input_shape = self.layers[layer_idx - 1].get_output_shape()
            if layer.has_weights == True:
                layer.initialize_params(optimizer=copy.deepcopy(self.optimizer))

    def add(self, layer: Layer) -> None:
        """Adds a layer to the network.

        Args:
            layer (Layer): An object of the Layer class which will be appended to the model.
        """
        self.layers.append(layer)

    def forward_pass(self, X: np.ndarray) -> np.ndarray:
        """Computes a forward pass of the network.

        Args:
            X (np.ndarray): The input to the network.

        Returns:
            np.ndarray: Output of the last layer of the network [shape: (batch_size x num_classes)].
        """
        prev_layer_output = X
        for layer in self.layers:
            prev_layer_output = layer.forward(prev_layer_output)
        return prev_layer_output

    def backward_pass(self, y_batch: np.ndarray, z: np.ndarray) -> None:
        """Computes a backward pass of the network (optimize).

        Args:
            y_batch (np.ndarray): The true labels.
            z (np.ndarray): The predicted labels.
        """
        # Traverse the layers in the reverse order
        # The gradient of the loss function [shape: (batch_size x num_classes)]
        grad = self.loss.objective_gradient(y_batch, z)
        for layer in self.layers[::-1]:
            grad = layer.backward(grad)

    def evaluate(self, X: np.ndarray, y: np.ndarray) -> None:
        """Evaluate the model and display all the required metrics (accuracy, r^2 score, etc.).

        Args:
            X (np.ndarray): The input dataset.
            y (np.ndarray): The true labels of the training data.
        """
        z = self.predict(X)

        for metric in self.metrics:
            self.history[metric.__name__].append(metric(y, z))

        text_to_display = ""
        for metric_name in self.history:
            text_to_display += f", {metric_name}: {self.history[metric_name][-1]}"
        print(text_to_display)

    def fit(self, dataset: Dataset, batch_size: int = 32, epochs: int = 50) -> None:
        """This method trains the model given the showml.utils.dataset.Dataset object (initialized with input data X and labels y).

        Args:
            dataset (Dataset): An object of the  showml.utils.dataset.Dataset class - the input dataset and true labels/values of the dataset.
            batch_size (int, optional): Number of samples per gradient update. Defaults to 32.
            epochs (int, optional): The number of epochs for training. Defaults to 50.
        """
        for epoch in range(1, epochs + 1):
            print(f"Epoch: {epoch}/{epochs}", end="")

            for X_batch, y_batch in generate_minibatches(
                dataset.X, dataset.y, batch_size, shuffle=True
            ):
                z = self.forward_pass(X_batch)
                self.backward_pass(y_batch, z)

            # Evaluate the model on the entire dataset
            self.evaluate(dataset.X, dataset.y)

    def predict(self, X: np.ndarray) -> np.ndarray:
        """Computes a forward pass of the model on the given data.

        Args:
            X (np.ndarray): The input data to the network.

        Returns:
            np.ndarray: Outputs of the last layer of the network [shape: (num_samples_of_X x num_classes)]].
        """
        return self.forward_pass(X)

    def summary(self) -> None:
        """Summarizes the model by displaying all layers, their parameters and total number of trainable parameters.
        """
        total_params = 0
        print(AsciiTable([[self.__class__.__name__]]).table)
        summary_data = [["Layer", "Params", "Output Shape"]]
        for layer in self.layers:
            summary_data.append(
                [
                    layer.__class__.__name__,
                    str(layer.get_params_count()),
                    str(layer.get_output_shape()),
                ]
            )
            total_params += layer.get_params_count()
        table = AsciiTable(summary_data)
        print(table.table)
        print(f"Total Params: {total_params}")



import numpy as np
from showml.utils.exceptions import DatasetSizeMismatchError, DataTypeError


class Dataset:
    """A Dataset class which is instantiated in order to validate the model inputs X and y before training/testing.

    Raises:
        DataTypeError: Raised when X and y have different lengths.
        DatasetSizeMismatchError: Raised when a variable has an unexpected datatype.
    """

    def __init__(self, X: np.ndarray, y: np.ndarray) -> None:
        """Constructor for the Dataset class.

        Args:
            X (np.ndarray): The input data.
            y (np.ndarray): The labels/values of the dataset.
        """
        self.X = X
        self.y = y
        self.validate_data_types()
        self.valid_dataset_length()

    def validate_data_types(self) -> None:
        """Validates if X and y are NumPy arrays (np.ndarray).
        """
        if not (isinstance(self.X, np.ndarray) and isinstance(self.y, np.ndarray)):
            raise DataTypeError("X and y must be NumPy arrays (np.ndarray)")

    def valid_dataset_length(self) -> None:
        """Validates if X and y have the same number of samples.
        """
        if self.X.shape[0] != self.y.shape[0]:
            raise DatasetSizeMismatchError(
                "Dimension 1 (number of samples) of X and y must the same"
            )



from showml.deep_learning.layers import Activation
import numpy as np


class Sigmoid(Activation):
    """A layer which applies the Sigmoid operation to an input.
    """

    def forward(self, X: np.ndarray) -> np.ndarray:
        return 1 / (1 + np.exp(-X))

    def backward(self, X: np.ndarray) -> np.ndarray:
        return self.forward(X) * (1 - self.forward(X))


class Relu(Activation):
    """A layer which applies the ReLU operation to an input.
    """

    def forward(self, X: np.ndarray) -> np.ndarray:
        return abs(X) * (X > 0)

    def backward(self, X: np.ndarray) -> np.ndarray:
        return 1.0 * (X > 0)


class Softmax(Activation):
    """A layer which applies the Softmax operation to an input.
    """

    def forward(self, X: np.ndarray) -> np.ndarray:
        e_x = np.exp(X - np.max(X, axis=-1, keepdims=True))
        return e_x / np.sum(e_x, axis=-1, keepdims=True)

    def backward(self, X: np.ndarray) -> np.ndarray:
        return self.forward(X) * (1 - self.forward(X))



import numpy as np
from showml.losses import MeanSquaredError, BinaryCrossEntropy
from showml.losses.loss_functions import CrossEntropy


def r2_score(y: np.ndarray, z: np.ndarray) -> float:
    """Calculate the r^2 (coefficient of determination) score of the model.

    Args:
        y (np.ndarray): The true values.
        z (np.ndarray): The predicted values.

    Returns:
        float: The r^2 score.
    """
    rss = np.sum(np.square(y - z))
    tss = np.sum(np.square(y - np.mean(y)))
    r_2 = 1 - (rss / tss)
    return r_2


def accuracy(y: np.ndarray, z: np.ndarray) -> float:
    """Compute the classification accuracy of the model.

    Args:
        y (np.ndarray): The true labels.
        z (np.ndarray): The predicted labels.

    Returns:
        float: The classification accuracy of the model.
    """
    if y.ndim == 1:
        # y and z are not one hot encoded
        true_class = y
        predicted_class = [1 if i > 0.5 else 0 for i in z]
    else:
        # y and z are one hot encoded
        true_class = np.argmax(y, axis=1)
        predicted_class = np.argmax(z, axis=1)

    return np.sum(true_class == predicted_class) / len(true_class)


def mean_squared_error(y: np.ndarray, z: np.ndarray) -> float:
    """Computes the Mean Squared Error (MSE).

    Args:
        y (np.ndarray): The true labels.
        z (np.ndarray): The predicted labels.

    Returns:
        float: The MSE value.
    """
    return MeanSquaredError().objective(y, z)


def binary_cross_entropy(y: np.ndarray, z: np.ndarray) -> float:
    """Computes the Binary Cross Entropy value (BCE).

    Args:
        y (np.ndarray): The true labels.
        z (np.ndarray): The predicted labels.

    Returns:
        float: the BCE value.
    """
    return BinaryCrossEntropy().objective(y, z)


def cross_entropy(y: np.ndarray, z: np.ndarray) -> float:
    """Computes the Cross Entropy value (CE).

    Args:
        y (np.ndarray): The true labels.
        z (np.ndarray): The predicted labels.

    Returns:
        float: the CE value.
    """
    return CrossEntropy().objective(y, z)



from typing import Tuple
import numpy as np
import pandas as pd
from typing import Tuple


def load_wine() -> Tuple[np.ndarray, np.ndarray]:
    """A method to load the Wine_Quality.csv file and initialize X_train and y_train.

    Returns:
        np.ndarray: The training data.
        np.ndarray: The training labels.
    """
    wine = (
        pd.read_csv(
            "./data/Wine_Quality.csv", sep=";", na_values="?", dtype={"ID": str}
        )
        .dropna()
        .reset_index()
    )
    X_train = wine[
        [
            # "fixed acidity",
            # "volatile acidity",
            # "citric acid",
            # "residual sugar",
            # "chlorides",
            # "free sulfur dioxide",
            # "total sulfur dioxide",
            "density",
            "pH",
            "sulphates",
            "alcohol",
        ]
    ].values

    # Remap wine quality to only 2 groups
    wine["quality"] = wine["quality"].map({3: 0, 4: 0, 5: 0, 6: 1, 7: 1, 8: 1, 9: 1})
    y_train = wine[["quality"]].values

    # Make y_train 1D if its not
    if y_train.ndim > 1:
        y_train = y_train[:, 0]
    return X_train, y_train


def load_auto() -> Tuple[np.ndarray, np.ndarray]:
    """A method to load the Auto.csv file and initialize X_train and y_train.

    Returns:
        np.ndarray: The training data.
        np.ndarray: The training labels.
    """
    auto = (
        pd.read_csv("./data/Auto.csv", na_values="?", dtype={"ID": str})
        .dropna()
        .reset_index()
    )
    X_train = auto[
        [
            "cylinders",
            "displacement",
            "horsepower",
            "weight",
            "acceleration",
            "year",
            "origin",
        ]
    ].values
    y_train = auto[["mpg"]].values

    # Make y_train 1D if its not
    if y_train.ndim > 1:
        y_train = y_train[:, 0]

    return X_train, y_train


def load_salary() -> Tuple[np.ndarray, np.ndarray]:
    """A method to load the Salary.csv file and initialize X_train and y_train.

    Returns:
        np.ndarray: The training data.
        np.ndarray: The training labels.
    """
    salary_data = (
        pd.read_csv("./data/Salary.csv", na_values="?", dtype={"ID": str})
        .dropna()
        .reset_index()
    )
    X_train = salary_data[["YearsExperience"]].values
    y_train = salary_data[["Salary"]].values

    # Make y_train 1D if its not
    if y_train.ndim > 1:
        y_train = y_train[:, 0]

    return X_train, y_train


def load_housepricedata() -> Tuple[np.ndarray, np.ndarray]:
    """A method to load the House_Price_Data.csv file and initialize X_train and y_train.

    Returns:
        np.ndarray: The training data.
        np.ndarray: The training labels.
    """
    house = pd.read_csv("./data/House_Price_Data.csv")

    X_train = house[
        [
            "LotArea",
            "OverallQual",
            "OverallCond",
            "TotalBsmtSF",
            "FullBath",
            "HalfBath",
            "BedroomAbvGr",
            "TotRmsAbvGrd",
            "Fireplaces",
            "GarageArea",
        ]
    ].values

    y_train = house["AboveMedianPrice"].values

    # Make y_train 1D if its not
    if y_train.ndim > 1:
        y_train = y_train[:, 0]
    return X_train, y_train



from typing import Tuple
from pydantic import BaseSettings


class GameWindowSettings(BaseSettings):
    # Window
    CELL_WIDTH: int = 9
    CELL_HEIGHT: int = 9
    CELL_MARGIN: int = 1
    FONT: str = "Arial"
    FONT_SIZE: int = 15
    CAPTION: str = "Conway's Game of Life - Cellular Automaton"


class Color:
    BLACK: Tuple[int, int, int] = (0, 0, 0)
    GRAY: Tuple[int, int, int] = (50, 50, 50)
    WHITE: Tuple[int, int, int] = (255, 255, 255)
